//=========================================================================
// rk.hh
//
// A collection of light-weight classes and utilities for calculations
// in relativistic kinematics. Boosts are implemented in such a way that
// particle masses are conserved exactly.
//
// I. Volobouev
// January 2008
//=========================================================================

#ifndef RK_RK_HH_
#define RK_RK_HH_

#include <utility>

#include "geom3.hh"

namespace rk {
    // Forward declarations
    class P4;
    class Boost;

    // The "P4" class can be used to represent time-like and light-like
    // vectors in 4-dimensional Minkowski space-time. The main purpose
    // of this class is to represent 4-momenta of relativistic particles.
    //
    // With some restrictions, it can also be used to represent space-like
    // vectors (e.g., results of intermediate 4-vector subtractions).
    // However, calling the "m()" method on a space-like vector will result
    // in a run-time error. This strongly limits usability of this code
    // for representing space-like vectors (for example, such vectors
    // can not be boosted). The advantage is that the mass can be memoized
    // for time-like and light-like vectors. This leads to a significant
    // speedup in a variety of calulations.
    //
    class P4
    {
    public:
        // The default constructor makes a 4-vector
        // with all components set to 0. Use with care.
        P4();

        // Constructor from 3-momentum and mass
        P4(const geom3::Vector3& p, double m, bool isEnergyNegative=false);

        // Do not use the following constructor
        // to make light-like vectors, use the one
        // with the explicit mass argument instead.
        P4(double e, const geom3::Vector3& p);

        // Basic accessors
        const geom3::Vector3& momentum() const;
        double e() const;
        double m() const;

        // Other useful accessors with obvious meaning
        double px() const;
        double py() const;
        double pz() const;
        double p() const;

        // Transverse momentum
        double pt() const;

        // Transverse energy
        double et() const;

        // Rapidity
        double rapidity() const;

        // Pseudorapidity
        double eta() const;

        // Various angular quantities. They are the same as
        // the corresponding functions of the 3-vectors.
        double theta() const;
        double cosTheta() const;
        double phi() const;

        // Speed and gamma-factor
        double beta() const;
        double gamma() const;

        // The following function is faster and has better precision
        // than the product beta()*gamma()
        double betaGamma() const;

        // 3-velocity in units of c
        geom3::Vector3 velocity() const;

        // 4-velocity
        P4 fourVelocity() const;

        // Scalar product : E1*E2 - p1.dot(p2)
        double dot(const P4&) const;

        // Boost this particle. If you want to get a new P4
        // object instead of boosting this one, use the "*"
        // operator of the boost object.
        P4& boost(const Boost& b);

        // Boost into the rest frame of this particle
        Boost restBoost() const;

        // Boost from the rest frame of this particle into the lab.
        // Equivalent to but slightly faster than restBoost().inverse().
        Boost labBoost() const;

        // Energy and momentum reversal. Use unary "-"
        // to reverse everything.
        P4 reverseE() const;
        P4 reverseP() const;

        // Unary operators
        P4 operator-() const;
        P4 operator+() const;

        // Binary operators
        friend bool operator==(const P4& l, const P4& r);
        friend bool operator!=(const P4& l, const P4& r);
        friend P4 operator+(const P4& l, const P4& r);
        friend P4 operator-(const P4& l, const P4& r);
        friend P4 operator*(const P4& l, const double& r);
        friend P4 operator*(const double& l, const P4& r);
        friend P4 operator/(const P4& l, const double& r);
        friend P4 operator*(const Boost& l, const P4& r);

        P4& operator+=(const P4& r);
        P4& operator-=(const P4& r);
        P4& operator*=(const double& d);
        P4& operator/=(const double& d);

    private:
        P4(double e, const geom3::Vector3& p, double m);
        P4(double e, const geom3::Vector3& p, bool nonNegative);
        P4(double e, const geom3::Vector3& p, double m, bool nonNegative);

        // The following function ensures m_ > 0.0 and causes
        // a run-time error if it is not
        void ensureMass_(void) const;

        geom3::Vector3 p_;
        double e_;
        mutable double m_;
        mutable bool msqIsNonNegative_;

        friend class Boost;
    };

    // The "Boost" class can be used to perform Lorentz transformations.
    // Here, the code assumes that the particle 4-momentum is known
    // in some system S and we need to know it in some system S' which
    // is moving w.r.t. S in the given direction with the speed
    // s = c*tanh(rapidity).
    //
    // The most convenient ways to create boosts are the "restBoost"
    // and "labBoost" methods of the "P4" class.
    //
    class Boost
    {
    public:
        // The default constructor builds a boost with 0 relative
        // velocity (unit transformation)
        Boost();

        explicit Boost(const geom3::Vector3& velocity_in_units_of_c);
        Boost(const geom3::UnitVector3& direction, double rapidity);

        // Basic accessors
        const geom3::UnitVector3& direction() const;
        double rapidity() const;

        // Inverse boost
        Boost inverse() const;

        // Boosts are performed by the "*" operator
        friend P4 operator*(const Boost& l, const P4& r);
        friend geom3::Vector3 operator*(const Boost& l,
                                        const geom3::Vector3& r);

        // The following operator can be used to figure out
        // how a direction changes under boost. Note that
        // the change in the vector magnitude is ignored.
        friend geom3::UnitVector3 operator*(const Boost& l,
                                            const geom3::UnitVector3& r);
        // Binary comparison operators
        friend bool operator==(const Boost& l, const Boost& r);
        friend bool operator!=(const Boost& l, const Boost& r);

    private:
        Boost(const geom3::UnitVector3& direction, double ch, double sh);

        geom3::UnitVector3 dir_;
        mutable double rapidity_;
        double c_;
        double s_;

        friend class P4;
    };

    // Invariant mass of 2 and 3 particles. More precise
    // (but slightly slower) than adding the 4-momenta
    // and then calling the m() function.
    double invMass(const P4& p1, const P4& p2);
    double invMass(const P4& p1, const P4& p2, const P4& p3);

    // Kinematic lambda function (in the literature on relativistic
    // kinematics it is usually denoted by the square root of greek
    // lambda). This function makes certain assumptions about its usage.
    // It has to be called with all non-negative arguments. Also,
    // the square root of one of the arguments must be larger than
    // the sum of the square roots of the two other.
    //
    // The center-of-mass momentum of the daughters in a two-body
    // decay is lambda(M_mother**2, M_dau1**2, M_dau2**2) / (2*M_mother).
    //
    double lambda(double x, double y, double z);

    // Random phase space particle decay into two daughters
    // with given masses. rnd0 and rnd1 should be random
    // numbers between 0 and 1. The first member of the pair
    // will have mass m1 and the second will have mass m2.
    // The mass of the parent must not be less than m1 + m2.
    std::pair<P4,P4> phaseSpaceDecay(const P4& parent, double m1, double m2,
                                     double rnd0, double rnd1);

    // The same function with a slightly different signature.
    // The results will be placed in *dau1 and *dau2.
    // Use the signature that is more appropriate for your code.
    // If your code can store the info in pairs of P4 objects and
    // can avoid unnecessary copying, the previous function will be
    // slightly faster to use.
    void phaseSpaceDecay(const P4& parent, double m1, double m2,
                         double rnd0, double rnd1, P4* dau1, P4* dau2);
}

std::ostream& operator<<(std::ostream& os, const rk::P4& v);

#include "rk_P4.icc"
#include "rk_Boost.icc"

#endif // RK_RK_HH_
